from __future__ import annotations

import dataclasses
import logging
import struct

from voice_stream.audio import AudioFormatError

logger = logging.getLogger(__name__)


def __generate_ogg_crc_lookup():
    polynomial = 0x04C11DB7
    crc_lookup = []

    for i in range(256):
        crc = i << 24
        for _ in range(8, 0, -1):
            if crc & 0x80000000:
                crc = (crc << 1) ^ polynomial
            else:
                crc = crc << 1
            crc = crc & 0xFFFFFFFF  # Ensure CRC remains a 32-bit value
        crc_lookup.append(crc)

    return crc_lookup


OGG_CRC_LOOKUP = __generate_ogg_crc_lookup()


@dataclasses.dataclass
class OggPage:
    data: bytes
    offset: int
    header_type: bytes
    granule: int
    serial_number: int
    page_sequence_number: int
    num_segments: bytes
    segment_lengths: list[int]
    header_length: int
    page_length: int

    @classmethod
    def from_data(cls, data: bytes, offset: int):
        # OGG page header structure:
        # 0-3 bytes: capture pattern (OggS)
        # 4 byte: stream structure version
        # 5 byte: header type flag
        # 6-13 bytes: granule position
        # 14-17 bytes: bitstream serial number
        # 18-21 bytes: page sequence number
        # 22-25 bytes: checksum (to be recalculated)
        # 26 byte: page segments
        if data[offset : offset + 4] != b"OggS":
            raise AudioFormatError(
                f"Invalid OGG page.  Expected 'OggS' header but got {data[offset:offset+4]}"
            )
        version = data[offset + 4]
        if version != 0:
            raise AudioFormatError(f"Last known OGG versions is 0.  {version} found.")
        header_type = data[offset + 5]
        granule = struct.unpack("<Q", data[offset + 6 : offset + 14])[0]
        serial_number = struct.unpack("<I", data[offset + 14 : offset + 18])[0]
        page_sequence_number = struct.unpack("<I", data[offset + 18 : offset + 22])[0]
        num_segments = data[offset + 26]
        segment_lengths = [data[offset + 27 + i] for i in range(num_segments)]
        header_length = 27 + num_segments
        page_length = header_length + sum(segment_lengths)
        return cls(
            data,
            offset,
            header_type,
            granule,
            serial_number,
            page_sequence_number,
            num_segments,
            segment_lengths,
            header_length,
            page_length,
        )

    def get_bytes(self) -> bytes:
        if self.offset == 0 and len(self.data) == self.page_length:
            return self.data
        else:
            return self.data[self.offset : self.offset + self.page_length]

    def update(
        self, new_header_type: bytes, granule_offset: int, sequence_offset: int
    ) -> OggPage:
        new_granule = granule_offset + self.granule
        granule_bytes = struct.pack("<Q", new_granule)
        new_sequence_number = sequence_offset + self.page_sequence_number
        sequence_bytes = struct.pack("<I", new_sequence_number)
        new_data = (
            self.data[self.offset : self.offset + 5]
            + struct.pack("<B", new_header_type)
            + granule_bytes
            + self.data[self.offset + 14 : self.offset + 18]
            + sequence_bytes
            + b"\x00\x00\x00\x00"
            + self.data[self.offset + 26 : self.offset + self.page_length]
        )
        checksum = self._compute_ogg_crc_checksum(new_data)
        checksum_bytes = struct.pack("<I", checksum)
        new_data = new_data[:22] + checksum_bytes + new_data[26:]
        return OggPage(
            new_data,
            0,
            new_header_type,
            new_granule,
            self.serial_number,
            new_sequence_number,
            self.num_segments,
            self.segment_lengths,
            self.header_length,
            self.page_length,
        )

    def log(self):
        checksum_matched = self.is_checksum_valid()
        logger.info(
            f"OGG page: offset={self.offset}, len={self.page_length}, header_type={self.header_type}, granule={self.granule}, serial_number={self.serial_number}, page_sequence_number={self.page_sequence_number}, checksum={checksum_matched}, num_segments={self.num_segments}, segment_lengths={self.segment_lengths}"
        )

    def is_checksum_valid(self):
        checksum = struct.unpack("<I", self.data[self.offset + 22 : self.offset + 26])[
            0
        ]
        return self.compute_checksum() == checksum

    def compute_checksum(self):
        checksum_buffer = (
            self.data[self.offset : self.offset + 22]
            + b"\x00\x00\x00\x00"
            + self.data[self.offset + 26 : self.offset + self.page_length]
        )
        return self._compute_ogg_crc_checksum(checksum_buffer)

    @classmethod
    def find_last_page(cls, data: bytes):
        return data.rfind(b"OggS")

    @classmethod
    def last_page_from_data(cls, data: bytes):
        position = data.rfind(b"OggS")
        return cls.from_data(data, position)

    @classmethod
    def is_full_page(cls, data: bytes, offset: int):
        """Determine if the data buffer contains the full OGG page.  It can contain extra bytes."""
        data_len = len(data)
        if data_len < offset + 27:
            return False
        num_segments = data[offset + 26]
        if data_len < offset + 27 + num_segments:
            return False
        segment_lengths = [data[offset + 27 + i] for i in range(num_segments)]
        header_length = 27 + num_segments
        page_length = header_length + sum(segment_lengths)
        return data_len >= offset + page_length

    @classmethod
    def _compute_ogg_crc_checksum(cls, data):
        checksum = 0
        for byte in data:
            tbl_idx = ((checksum >> 24) ^ byte) & 0xFF
            checksum = (OGG_CRC_LOOKUP[tbl_idx] ^ (checksum << 8)) & 0xFFFFFFFF

        return checksum


@dataclasses.dataclass
class OpusIdPacket:
    data: bytes
    offset: int
    output_channel_count: bytes
    pre_skip: int
    input_sample_rate: int
    output_gain: int
    channel_mapping_family: bytes

    @classmethod
    def is_opus_encoded(cls, data: bytes, offset: int) -> bool:
        return data[offset : offset + 8] == b"OpusHead"

    @classmethod
    def from_ogg_page(cls, page: OggPage) -> OpusIdPacket:
        return cls.from_data(page.data, page.offset + page.header_length)

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> OpusIdPacket:
        header = data[offset : offset + 8]
        if header != b"OpusHead":
            raise AudioFormatError(
                f"Invalid Opus ID packet header '{header}'.  Should be 'OpusHead'."
            )
        version = data[offset + 8]
        if version != 1:
            raise AudioFormatError(f"Last known OPUS versions is 1.  {version} found.")
        output_channel_count = data[offset + 9]
        pre_skip = struct.unpack("<H", data[offset + 10 : offset + 12])[0]
        input_sample_rate = struct.unpack("<I", data[offset + 12 : offset + 16])[0]
        output_gain = struct.unpack("<H", data[offset + 16 : offset + 18])[0]
        channel_mapping_family = data[offset + 18]
        assert (
            channel_mapping_family == 0
        ), f"Channel mapping is not yet supported.  Channel mapping is {channel_mapping_family}"
        # if channel_mapping_family != 0:
        #     stream_count = data[offset+20]
        #     coupled_count = data[offset+21]
        #     channel_mapping = [data[offset+22+i] for i in range(output_channel_count)]
        return cls(
            data,
            offset,
            output_channel_count,
            pre_skip,
            input_sample_rate,
            output_gain,
            channel_mapping_family,
        )

    def log(self):
        logger.info(
            f"Opus ID packet: offset={self.offset}, output_channel_count={self.output_channel_count}, pre_skip={self.pre_skip}, input_sample_rate={self.input_sample_rate}, output_gain={self.output_gain}, channel_mapping_family={self.channel_mapping_family}"
        )


@dataclasses.dataclass
class OpusCommentPacket:
    data: bytes
    offset: int
    vendor_string: bytes
    user_comment_list: list[bytes]

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> OpusCommentPacket:
        header = data[offset : offset + 8]
        assert (
            header == b"OpusTags"
        ), f"Invalid Opus comment header '{header}'.  Should be 'OpusTags'."
        vendor_length = struct.unpack("<I", data[offset + 8 : offset + 12])[0]
        vendor_string = data[offset + 12 : offset + 12 + vendor_length]
        user_comment_list_length = struct.unpack(
            "<I", data[offset + 12 + vendor_length : offset + 16 + vendor_length]
        )[0]
        user_comment_list = []
        for i in range(user_comment_list_length):
            comment_length = struct.unpack(
                "<I",
                data[
                    offset
                    + 16
                    + vendor_length
                    + i * 4 : offset
                    + 20
                    + vendor_length
                    + i * 4
                ],
            )[0]
            comment_string = data[
                offset
                + 20
                + vendor_length
                + i * 4 : offset
                + 20
                + vendor_length
                + i * 4
                + comment_length
            ]
            user_comment_list.append(comment_string)
        return cls(data, offset, vendor_string, user_comment_list)

    def log(self):
        logger.info(
            f"Opus comment header: offset={self.offset}, vendor_string={self.vendor_string}, user_comment_list={self.user_comment_list}"
        )



import asyncio
from typing import AsyncIterator


class QueueWithException:
    """A queue that propagates exceptions.
    If a None is enqueued and the exception is set, then the exception is thrown on put.
    """

    def __init__(self):
        self.queue = asyncio.Queue()
        self.exception = None

    async def enqueue_iterator(self, async_iter: AsyncIterator):
        if not hasattr(async_iter, "__aiter__"):
            raise ValueError(f"Object {async_iter} is not an async iterator")
        try:
            async for item in async_iter:
                await self.queue.put(item)
        except Exception as e:
            self.exception = e
        # Signal end of iteration
        await self.queue.put(None)

    def qsize(self):
        return self.queue.qsize()

    async def put(self, item):
        await self.queue.put(item)

    async def get(self):
        ret = await self.queue.get()
        if ret is None and self.exception:
            exception = self.exception
            self.exception = None
            raise exception
        return ret

    def get_nowait(self):
        ret = self.queue.get_nowait()
        if ret is None and self.exception:
            exception = self.exception
            self.exception = None
            raise exception
        return ret

    def set_exception(self, exception: Exception):
        self.exception = exception



import json
import logging
from asyncio import CancelledError
from typing import AsyncIterator

from hypercorn.utils import UnexpectedMessageError
from quart import websocket

logger = logging.getLogger(__name__)


async def quart_websocket_source() -> AsyncIterator[str]:
    """Creates an ASyncIterator for messages coming from the Quart websocket object."""
    while True:
        try:
            message = await websocket.receive()
        except CancelledError:
            logger.debug("websocket.receive is cancelled.")
            return
        yield message


async def quart_websocket_sink(async_iter: AsyncIterator[str]) -> None:
    """Takes an async iterator and sends everything to a Quart websocket.  Handles strings, dicts, or bytes."""
    try:
        async for message in async_iter:
            if isinstance(message, dict):
                message = json.dumps(message)
            await websocket.send(message)
    except UnexpectedMessageError as e:
        if "ASGIWebsocketState.CLOSED" in str(e):
            logger.info("Websocket closed before all queued messages were sent.")
        else:
            raise e



import asyncio
import inspect
import logging
import string
import time
from typing import AsyncIterator, Tuple, Callable, Union

from pydantic import BaseModel

from voice_stream.audio import AudioFormat, AudioFormatError
from voice_stream.audio.audio_streams import (
    get_audio_length,
    ogg_page_separator_step,
    mp3_chunk_step,
)
from voice_stream.basic_streams import (
    chunk_bytes_step,
    fork_step,
    map_step,
    extract_value_step,
    map_future,
    FutureOrObj,
    resolve_obj_or_future,
    async_init_step,
    str_buffer_step,
    filter_step,
)
from voice_stream.events import TimedText
from voice_stream.substreams import substream_step

logger = logging.getLogger(__name__)


class AudioWithText(BaseModel):
    audio: bytes
    text: str
    audio_format: AudioFormat


TextToSpeechStep = Callable[[AsyncIterator[str]], AsyncIterator[AudioWithText]]


def buffer_tts_text_step(async_iter: AsyncIterator[str]):
    pipe = wait_for_punctuation_step(async_iter)
    pipe = str_buffer_step(pipe)
    pipe = filter_step(pipe, lambda x: len(x) > 0)
    return pipe


def tts_with_buffer_and_rate_limit_step(
    async_iter: AsyncIterator[str], tts_step, buffer_seconds: float = 0.5
) -> (AsyncIterator[bytes], AsyncIterator[str]):
    pipe = buffer_tts_text_step(async_iter)
    pipe = tts_step(pipe)
    pipe, audio_format_f = extract_value_step(pipe, lambda x: x.audio_format)
    pipe, text_output = tts_rate_limit_step(
        pipe, audio_format_f, buffer_seconds=buffer_seconds
    )
    return pipe, text_output


async def wait_for_punctuation_step(
    async_iter: AsyncIterator[str],
) -> AsyncIterator[str]:
    """For TTS, we want to generate as soon as possible, but we don't want to do partial phrases.  Go until we get
    a punctuation mark or an empty token (which signals the end of a phrase)."""
    buffer = ""
    async for item in async_iter:
        buffer += item
        has_punc = any(char in string.punctuation for char in item.strip())
        if len(item) == 0 or has_punc:
            yield buffer
            buffer = ""


def tts_rate_limit_step(
    async_iter: AsyncIterator[AudioWithText],
    audio_format: FutureOrObj[AudioFormat],
    buffer_seconds: float = 0.5,
    include_text_output: bool = True,
) -> Union[AsyncIterator[bytes], Tuple[AsyncIterator[bytes], AsyncIterator[str]]]:
    """Breaks a single TextToSpeechOutput into smaller chunks."""

    async def tts_output_to_timed_text(tts: AudioWithText) -> TimedText:
        resolved_audio_format = await resolve_obj_or_future(audio_format)
        duration = get_audio_length(resolved_audio_format, tts.audio)
        return TimedText(tts.text, duration_in_seconds=duration)

    if include_text_output:
        audio, text = fork_step(async_iter, backpressure=True)
        text = map_step(text, tts_output_to_timed_text)
        # text = log_step(text, "Timed text in")
        text = timed_text_rate_limit_step(text, buffer_seconds=buffer_seconds)
        # text = log_step(text, "Timed text out")
    else:
        audio = async_iter

    audio = map_step(audio, lambda x: x.audio)
    # audio = log_step(audio, "Audio out", lambda x: len(x))
    audio = audio_rate_limit_step(audio, audio_format, buffer_seconds=buffer_seconds)

    if include_text_output:
        return audio, text
    else:
        return audio


def audio_rate_limit_step(
    async_iter: AsyncIterator[bytes],
    audio_format: FutureOrObj[AudioFormat],
    buffer_seconds: float,
):
    def init(async_iter, af):
        if af == AudioFormat.WAV_MULAW_8KHZ:
            SAMPLE_RATE = 8000
            audio = chunk_bytes_step(async_iter, int(buffer_seconds * SAMPLE_RATE))
            audio = raw_audio_rate_limit_step(
                audio, SAMPLE_RATE, buffer_seconds=buffer_seconds
            )
        elif af == AudioFormat.OGG_OPUS:
            audio = _opus_rate_limit_step(async_iter, buffer_seconds=buffer_seconds)
        elif af == AudioFormat.MP3:
            audio = _mp3_rate_limit_step(async_iter, buffer_seconds=buffer_seconds)
        else:
            raise AudioFormatError(f"Unsupported audio format: {audio_format}")
        return audio

    async def async_init(async_iter):
        af = await resolve_obj_or_future(audio_format)
        return init(async_iter, af)

    if inspect.isawaitable(audio_format):
        return async_init_step(async_iter, async_init)
    else:
        return init(async_iter, audio_format)


def _opus_rate_limit_step(async_iter: AsyncIterator[bytes], buffer_seconds: float):
    def get_bytes_per_second(data):
        duration = get_audio_length(AudioFormat.OGG_OPUS, data)
        length_in_bytes = len(data)
        return int(length_in_bytes / duration)

    def opus_substream(async_iter: AsyncIterator[bytes]):
        pipe, bytes_per_second_f = extract_value_step(async_iter, get_bytes_per_second)
        chunk_size_f = map_future(
            bytes_per_second_f,
            lambda bytes_per_second: int(buffer_seconds * bytes_per_second / 2),
        )
        pipe = chunk_bytes_step(pipe, chunk_size_f)
        # pipe = log_step(pipe, "Audio Chunk", lambda x: len(x))
        pipe = raw_audio_rate_limit_step(
            pipe, bytes_per_second_f, buffer_seconds=buffer_seconds
        )
        pipe = ogg_page_separator_step(pipe)
        # pipe = log_step(pipe, "Full OGG Page", lambda x: len(x))
        return pipe

    ret = substream_step(async_iter, opus_substream)
    return ret


def _mp3_rate_limit_step(async_iter: AsyncIterator[bytes], buffer_seconds: float):
    def get_bytes_per_second(data):
        seconds = get_audio_length(AudioFormat.MP3, data)
        length_in_bytes = len(data)
        return int(length_in_bytes / seconds)

    def mp3_substream(async_iter: AsyncIterator[bytes]):
        pipe, bytes_per_second_f = extract_value_step(async_iter, get_bytes_per_second)
        chunk_size_f = map_future(
            bytes_per_second_f,
            lambda bytes_per_second: int(buffer_seconds * bytes_per_second / 2),
        )
        pipe = mp3_chunk_step(pipe, chunk_size_f)
        pipe = raw_audio_rate_limit_step(
            pipe, bytes_per_second_f, buffer_seconds=buffer_seconds
        )
        return pipe

    ret = substream_step(async_iter, mp3_substream)
    return ret


async def raw_audio_rate_limit_step(
    async_iter: AsyncIterator[bytes],
    bytes_per_second: FutureOrObj[int],
    buffer_seconds: FutureOrObj[float],
) -> AsyncIterator[bytes]:
    """Limits the rate of sending audio bytes down the pipe.  Note that this step always sends a full chunk.
    buffer_seconds is the number of seconds of audio left before we send the next chunk.  Usually, you will want to put
    a max size chunk step in before this.  buffer_seconds should be large enough to make sure the send executes.
    """
    last_send = time.perf_counter()
    queued_audio_seconds = 0

    def compute_remaining(now):
        # Compute the amount of audio remaining to be played in seconds.
        return max(0, queued_audio_seconds - (now - last_send))

    async for item in async_iter:
        resolved_buffer_seconds = await resolve_obj_or_future(buffer_seconds)
        now = time.perf_counter()
        # Compute the amount of audio remaining to be played in seconds.
        remaining_audio_seconds = compute_remaining(now)
        # If we have more than the buffer, sleep until we hit the buffer limit
        if remaining_audio_seconds > resolved_buffer_seconds:
            await asyncio.sleep(remaining_audio_seconds - resolved_buffer_seconds)
            # We don't know how long we actually slept.
            now = time.perf_counter()
            remaining_audio_seconds = compute_remaining(now)
        resolved_bytes_per_second = await resolve_obj_or_future(bytes_per_second)
        queued_audio_seconds = (
            remaining_audio_seconds + len(item) / resolved_bytes_per_second
        )
        last_send = now
        yield item


async def timed_text_rate_limit_step(
    async_iter: AsyncIterator[TimedText], buffer_seconds: float = 0.5
) -> AsyncIterator[str]:
    """Used in conjunction with text to speech.  Takes in timed text and outputs it token by token given the time.
    Assumes all tokens take equal time.
    """
    last_send = time.perf_counter()
    queued_text_seconds = 0

    def compute_remaining(now):
        # Compute the amount of text remaining to be played in seconds.
        return max(0, queued_text_seconds - (now - last_send))

    async def break_into_tokens(timed_text: TimedText):
        tokens = timed_text.text.split(" ")
        if len(tokens) > 0:
            token_duration = timed_text.duration_in_seconds / len(tokens)
            tokens = [tokens[0]] + [" " + _ for _ in tokens[1:]]
            for token in tokens:
                yield TimedText(token, token_duration)
        else:
            yield timed_text

    async for timed_text in async_iter:
        async for token in break_into_tokens(timed_text):
            now = time.perf_counter()
            remaining_text_seconds = compute_remaining(now)
            # If we have more than the buffer, sleep until we hit the buffer limit
            if remaining_text_seconds > buffer_seconds:
                await asyncio.sleep(remaining_text_seconds - buffer_seconds)
                # We don't know how long we actually slept.
                now = time.perf_counter()
                remaining_text_seconds = compute_remaining(now)
            queued_text_seconds = remaining_text_seconds + token.duration_in_seconds
            last_send = now
            yield token.text


async def min_tokens_step(
    async_iter: AsyncIterator[str], min_tokens: int = 3
) -> AsyncIterator[str]:
    """Makes sure we have at least minimum number of tokens, but always flushes on an empty token."""
    buffer = ""
    num_tokens = 0
    async for item in async_iter:
        if len(item) == 0:
            if len(buffer) > 0:
                yield buffer
                buffer = ""
        else:
            buffer += item
            num_tokens += 1
            if num_tokens >= min_tokens:
                yield buffer
                buffer = ""
                num_tokens = 0
    if len(buffer) > 0:
        yield buffer



from __future__ import annotations

from typing import AsyncIterator, Callable, Any, Optional

from langchain_core.runnables import Runnable

from voice_stream.basic_streams import (
    fork_step,
)
from voice_stream.speech_to_text_streams import (
    filter_spurious_speech_start_events_step,
    SpeechStep,
)
from voice_stream.substreams import cancelable_substream_step
from voice_stream.text_to_speech_streams import (
    TextToSpeechStep,
    tts_with_buffer_and_rate_limit_step,
)

GenericStepFunc = Callable[[AsyncIterator[Any]], AsyncIterator[Any]]


class LangchainVoiceFlow:
    def __init__(
        self,
        text_input: AsyncIterator[str],
        text_output: AsyncIterator[str],
        audio_output: AsyncIterator[bytes],
    ):
        self.text_input = text_input
        self.text_output = text_output
        self.audio_output = audio_output

    @classmethod
    def create(
        cls,
        audio_input: AsyncIterator[bytes],
        speech_step: SpeechStep,
        chain: Runnable,
        tts_step: TextToSpeechStep,
        langchain_postprocess: Optional[GenericStepFunc] = None,
    ) -> LangchainVoiceFlow:
        pipe, speech_start = speech_step(audio_input)
        speech_start = filter_spurious_speech_start_events_step(speech_start)
        pipe, text_input = fork_step(pipe)

        def create_output_chain(
            pipe: AsyncIterator[str],
        ) -> (AsyncIterator[bytes], AsyncIterator[str]):
            pipe = langchain_stream_step(pipe, chain)
            if langchain_postprocess:
                pipe = langchain_postprocess(pipe)
            # pipe = log_step(pipe, "Token")
            return tts_with_buffer_and_rate_limit_step(pipe, tts_step)

        pipe, text_output = cancelable_substream_step(
            pipe, speech_start, create_output_chain
        )
        return cls(
            text_input=text_input,
            text_output=text_output,
            audio_output=pipe,
        )


async def langchain_stream_step(
    async_iter: AsyncIterator[str], chain: Runnable[str, str]
) -> AsyncIterator[str]:
    """Runs a chain for each text item sent in, streams back response tokens."""
    async for text in async_iter:
        async for token in chain.astream(text):
            yield token
        yield ""  # Forces an empty token to help flush buffers.
    # Note on cancelling - https://github.com/langchain-ai/langchain/issues/11959



import logging
from io import BytesIO

from mutagen.mp3 import MP3

logger = logging.getLogger(__name__)

# Bitrates and sampling rates values for different versions and layers
MP3_BITRATES = {
    "V1L1": [None, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
    "V1L2": [None, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384],
    "V1L3": [None, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320],
    "V2L1": [None, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
    "V2L2": [None, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
    "V2L3": [None, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
}

MP3_SAMPLING_RATES = {
    "V1": [44100, 48000, 32000],
    "V2": [22050, 24000, 16000],
    "V2.5": [11025, 12000, 8000],
}


def find_frame_boundaries(mp3_data):
    # Function to find frame boundaries
    # This is a simplified version and might need more robust handling
    info = MP3(BytesIO(mp3_data)).info
    boundaries = []
    index = 0
    while index < len(mp3_data):
        if mp3_data[index] == 0xFF and (mp3_data[index + 1] & 0xE0) == 0xE0:
            boundaries.append(index)
            # Simplified frame size calculation (actual calculation is more complex)
            frame_size = calculate_frame_size(
                mp3_data[index : index + 4], version=info.version, layer=info.layer
            )
            index += frame_size
        else:
            logger.warning(f"Invalid frame header at index {index}")
            index += 1
    return boundaries


def calculate_frame_size(header, version, layer):
    # Extracting information from the header
    bitrate_index = (header[2] >> 4) & 0x0F
    sampling_rate_index = (header[2] >> 2) & 0x03
    padding = (header[2] >> 1) & 0x01

    # Look up the bitrate and sampling rate
    key = f"V{version}L{layer}"
    bitrate = MP3_BITRATES[key][bitrate_index] * 1000
    sampling_rate = MP3_SAMPLING_RATES[f"V{version}"][sampling_rate_index]

    # Calculate frame size
    frame_size = 144 * bitrate // sampling_rate + padding
    return frame_size


def calculate_split_points(frame_boundaries, buffer_size, max_size):
    """
    Calculate the split points for chunking an MP3 buffer.

    Each chunk will be no larger than max_size. If a single frame is larger than max_size,
    it will be a chunk on its own.

    :param frame_boundaries: A list of integers representing the start of each MP3 frame.
    :param buffer_size: The total size of the MP3 buffer.
    :param max_size: The maximum size of each chunk.
    :return: A list of integers representing the split points in the buffer.
    """
    split_points = []
    current_chunk_start = 0

    for i, boundary in enumerate(frame_boundaries):
        next_boundary = (
            frame_boundaries[i + 1] if i + 1 < len(frame_boundaries) else buffer_size
        )
        frame_size = next_boundary - boundary

        if frame_size > max_size:
            # If a single frame is larger than max_size, it becomes its own chunk
            if current_chunk_start < boundary:
                split_points.append(boundary)
            split_points.append(next_boundary)
            current_chunk_start = next_boundary
        elif boundary - current_chunk_start + frame_size > max_size:
            # Start a new chunk
            split_points.append(boundary)
            current_chunk_start = boundary

    # Add the last boundary if there's remaining data
    if current_chunk_start < buffer_size:
        split_points.append(buffer_size)

    return split_points



from voice_stream.audio.audio import AudioFormat, AudioFormatError
from voice_stream.audio.audio_streams import (
    ogg_page_separator_step,
    ogg_concatenator_step,
    wav_mulaw_file_source,
    wav_mulaw_file_sink,
)

__all__ = [
    AudioFormat,
    AudioFormatError,
    ogg_page_separator_step,
    ogg_concatenator_step,
    wav_mulaw_file_source,
    wav_mulaw_file_sink,
]



import asyncio
import inspect
import json
import logging
from asyncio import Future, QueueEmpty
from typing import (
    AsyncIterator,
    Any,
    Tuple,
    Callable,
    Union,
    TypeVar,
    Optional,
    Type,
    Iterable,
    Coroutine,
    List,
)

import aiofiles

from voice_stream._queue_with_exception import QueueWithException

T = TypeVar("T")
Output = TypeVar("Output")
FutureOrObj = Union[T, asyncio.Future[T]]

logger = logging.getLogger(__name__)

# Streams are an abstraction on top of ASyncIterators.
# Sources are iterators.
# Sinks take an iterator.
# Steps take iterators and return iterators


def to_tuple(obj) -> Tuple:
    if isinstance(obj, Tuple):
        return obj
    if isinstance(obj, List):
        return tuple(obj)
    else:
        return (obj,)


def from_tuple(obj):
    return obj[0] if len(obj) == 1 else obj


def is_async_iterator(obj):
    return inspect.iscoroutinefunction(getattr(obj, "__anext__", None)) and hasattr(
        obj, "__aiter__"
    )


def map_future(f: Future[T], func: Callable[[T], Output]) -> Future[Output]:
    """Returns a future that will be resolved with the result of applying func to the result of f.

    Words are made lowercase and punctuation is removed
    before counting.

    Parameters
    ----------
    input_file : str
        Path to text file.

    Returns
    -------
    collections.Counter
        dict-like object where keys are words and values are counts.

    Examples
    --------
    >>> count_words("text.txt")
    """
    loop = asyncio.get_running_loop()
    ret = loop.create_future()

    def callback(fut: Future[T]) -> None:
        try:
            ret.set_result(func(fut.result()))
        except Exception as e:
            ret.set_exception(e)

    f.add_done_callback(callback)
    return ret


async def resolve_obj_or_future(obj: FutureOrObj[T]) -> T:
    """Returns the result of an object or a future"""
    if isinstance(obj, asyncio.Future):
        return await obj
    return obj


def empty_source() -> AsyncIterator[T]:
    """Returns an empty async iterator that immediately sends an end of iteration."""
    return array_source([])


async def empty_sink(async_iter: AsyncIterator[T]) -> None:
    """An async iterator created from an array.  Returns the array."""
    async for _ in async_iter:
        pass
    # logger.info("Empty sink finished")


async def single_source(item: T) -> AsyncIterator[T]:
    """An async iterator created from an array."""
    await asyncio.sleep(0)
    yield item


async def array_source(array: list[T]) -> AsyncIterator[T]:
    """An async iterator created from an array."""
    for item in array:
        await asyncio.sleep(0)
        yield item


async def array_sink(async_iter: AsyncIterator[T]) -> list[T]:
    """An async iterator created from an array.  Returns the array."""
    array = []
    async for item in async_iter:
        array.append(item)
    return array


class QueueAsyncIterator:
    """An AsyncIterator that works over its own queue and adds a 'put' method."""

    def __init__(self):
        self.queue = asyncio.Queue()
        self.iter = _make_queue_iterator(self.queue)

    async def put(self, item):
        return await self.queue.put(item)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self.iter.__anext__()


def queue_source(queue: FutureOrObj[asyncio.Queue[T]] = None) -> AsyncIterator[T]:
    """Returns an async iterator over objects in a queue.  The queue must be closed by putting None into it."""
    if queue:

        async def gen():
            resolved_queue = await resolve_obj_or_future(queue)
            while True:
                try:
                    item = await resolved_queue.get()
                    if item is None:
                        break
                except asyncio.CancelledError:
                    # logger.debug("Queue iterator cancelled.")
                    break
                yield item

        return gen()
    else:
        return QueueAsyncIterator()


async def queue_sink(
    async_iter: AsyncIterator[T],
    queue: Optional[FutureOrObj[Union[QueueWithException, asyncio.Queue]]] = None,
) -> asyncio.Queue:
    """Writes each element of the async_iter to a queue"""
    resolved_queue = None
    try:
        async for message in async_iter:
            if resolved_queue is None:
                if queue:
                    resolved_queue = await resolve_obj_or_future(queue)
                else:
                    resolved_queue = asyncio.Queue()
            await resolved_queue.put(message)
    except Exception as e:
        if resolved_queue and hasattr(resolved_queue, "set_exception"):
            await resolved_queue.set_exception(e)
        else:
            raise e
    if resolved_queue:
        await resolved_queue.put(None)  # signal completion
    return resolved_queue


async def text_file_source(filename: str) -> AsyncIterator[str]:
    """Returns an async iterator over the lines in the file"""
    async with aiofiles.open(filename, "rt") as f:
        async for line in f:
            line = line[:-1] if line[-1] == "\n" else line
            yield line


async def text_file_sink(
    async_iter: AsyncIterator[Any], filename: FutureOrObj[str]
) -> None:
    """Write each element of the async_iter to a file as a line"""
    return await _file_sink(async_iter, filename, "wt", lambda x: f"{x}\n")


async def binary_file_source(
    filename: str, chunk_size: int = 4096
) -> AsyncIterator[bytes]:
    """Returns an async iterator over the bytes in the file"""
    async with aiofiles.open(filename, "rb") as f:
        while True:
            chunk = await f.read(chunk_size)
            if not chunk:
                break
            yield chunk


async def binary_file_sink(
    async_iter: AsyncIterator[bytes], filename: FutureOrObj[str]
) -> None:
    """Append each block of bytes from async_iter to the file"""
    return await _file_sink(async_iter, filename, "wb", lambda x: x)


async def _file_sink(
    async_iter: AsyncIterator[T],
    filename: FutureOrObj[str],
    mode: str,
    prep_func: Callable[[T], Any],
) -> None:
    f = None
    try:
        async for block in async_iter:
            # We wait to resolve the filename until we have a message to write
            if f is None:
                filename = await resolve_obj_or_future(filename)
                f = await aiofiles.open(filename, mode)
            await f.write(prep_func(block))
    finally:
        if f is not None:
            await f.close()
            logger.debug(f"Closed {filename}")


def map_str_to_json_step(async_iter: AsyncIterator[str]) -> AsyncIterator[dict]:
    return map_step(async_iter, lambda x: json.loads(x))


async def flatten_step(async_iter: AsyncIterator[Iterable[T]]) -> AsyncIterator[T]:
    """Takes an async iterator where each item is iterable and iterates over it."""
    async for item in async_iter:
        for subitem in item:
            yield subitem


async def async_flatten_step(
    async_iter: AsyncIterator[AsyncIterator[T]],
) -> AsyncIterator[T]:
    """Takes an async iterator where each item is an AsyncIterator and iterates over it."""
    async for item in async_iter:
        async for subitem in item:
            yield subitem


def async_init_step(
    async_iter: AsyncIterator[T],
    f: Callable[[AsyncIterator[T]], Coroutine],
    num_outputs: int = 1,
) -> AsyncIterator[Output]:
    class AsyncInitInputIterator:
        """Wraps the input iterator and eagerly gets the first item.  This pulls data through the stream in case our
        async_init function is waiting on it."""

        def __init__(self):
            self.iter = async_iter.__aiter__()
            self.first_item_task = asyncio.create_task(self.iter.__anext__())

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.first_item_task:
                ret = self.first_item_task
                self.first_item_task = None
                return await ret
            else:
                return await self.iter.__anext__()

    class AsyncInitOutputIterator:
        """Wraps the output and doesn't create the real iterator until the first item is requested."""

        def __init__(self, init_event):
            self.iter = None
            self.init_event = init_event

        def __aiter__(self):
            return self

        async def __anext__(self):
            if not self.iter:
                await self.init_event.wait()
            return await self.iter.__anext__()

    init_event = asyncio.Event()
    outputs = [AsyncInitOutputIterator(init_event) for _ in range(num_outputs)]

    async def do_init():
        input = AsyncInitInputIterator()
        step = await f(input)

        for output, iterable in zip(outputs, to_tuple(step)):
            output.iter = iterable.__aiter__()
        init_event.set()

    asyncio.create_task(do_init())

    if len(outputs) == 1:
        return outputs[0]
    else:
        return outputs


def extract_value_step(
    async_iter: AsyncIterator[T],
    value: Callable[[T], Any],
    condition: Optional[Callable[[T], bool]] = None,
) -> Tuple[AsyncIterator[T], asyncio.Future[Any]]:
    """Extracts a value from the iterator.  Returns the iterator and a future containing the value"""
    loop = asyncio.get_running_loop()
    fut = loop.create_future()

    async def chain():
        async for item in async_iter:
            if not fut.done() and (condition is None or condition(item)):
                fut.set_result(value(item))
            yield item

    return chain(), fut


async def exception_handler_step(
    async_iter: AsyncIterator[T],
    exception_type: Type[BaseException],
    exception_handler: Callable[[BaseException], None],
) -> AsyncIterator[T]:
    """Wraps an async iterator and logs any exceptions that occur."""
    try:
        async for item in async_iter:
            yield item
    except exception_type as e:
        exception_handler(e)


async def log_step(
    async_iter: AsyncIterator[T], name: str, formatter: Callable[[T], Any] = lambda x: x
) -> AsyncIterator[T]:
    """Logs out the messages coming through a source."""
    async for item in async_iter:
        formatted = formatter(item)
        logger.info(f"{name}: {formatted}")
        yield item


async def count_step(async_iter: AsyncIterator[T], name: str) -> AsyncIterator[T]:
    """Counts the number of messages streamed through a pipe."""
    counter = 0
    async for item in async_iter:
        counter += 1
        yield item
    logger.info(f"{name} count: {counter}")


async def filter_step(
    async_iter: AsyncIterator[T], condition: Callable[[T], bool]
) -> AsyncIterator[T]:
    async for item in async_iter:
        if condition(item):
            yield item


async def map_step(
    async_iter: AsyncIterator[T],
    func: Callable[[T], Output],
    ignore_none: Optional[bool] = False,
) -> AsyncIterator[Output]:
    """Handles both sync and async functions
    Optional 'ignore_none' clause indicates that a value of None should be dropped.  Allows filtering and mapping in one step.
    """
    is_async = inspect.iscoroutinefunction(func)
    async for item in async_iter:
        v = func(item)
        if is_async:
            v = await v
        if (not ignore_none) or v:
            yield v


async def concat_step(*async_iters: List[AsyncIterator[T]]) -> AsyncIterator[T]:
    """Concatenates multiple iterators into a single iterator."""
    for async_iter in async_iters:
        async for item in async_iter:
            yield item


async def chunk_bytes_step(
    async_iter: AsyncIterator[bytes], chunk_size: FutureOrObj[int]
) -> AsyncIterator[bytes]:
    """Breaks byte buffers into chunks with a maximum size."""
    async for item in async_iter:
        resolved_chunk_size = await resolve_obj_or_future(chunk_size)
        for i in range(0, len(item), resolved_chunk_size):
            data = item[i : i + resolved_chunk_size]
            # logger.debug(f"Chunked {len(data)} bytes")
            yield data


async def min_size_bytes_step(
    async_iter: AsyncIterator[bytes], min_size: FutureOrObj[int]
) -> AsyncIterator[bytes]:
    """Ensures byte streams have a least a minimum size."""
    buffer = b""
    async for item in async_iter:
        combined = buffer + item
        if len(combined) >= await resolve_obj_or_future(min_size):
            yield combined
            buffer = b""
        else:
            buffer = combined
    if len(buffer) > 0:
        yield buffer


async def merge_step(*async_iters: list[AsyncIterator[T]]) -> AsyncIterator[T]:
    """Merges multiple iterators into one.  Differs from concat in that it takes from both iterators at the same time.
    An exception in any iterators cancels all."""
    queue = QueueWithException()
    tasks = [asyncio.create_task(queue.enqueue_iterator(it)) for it in async_iters]

    try:
        completed = 0
        while completed < len(async_iters):
            item = await queue.get()
            if item is None:
                completed += 1
            else:
                yield item
    finally:
        for task in tasks:
            task.cancel()  # Cancel all tasks


async def merge_as_dict_step(
    dict_iters: dict[str, AsyncIterator[Any]]
) -> AsyncIterator[dict]:
    """Takes several iterators and collects each into different keys in a dictionary.  The first in the list driver the output."""
    # logger.debug(f"merge_as_dict_step")
    if not dict_iters:
        raise ValueError("merge_as_dict_step requires at least on iterator")
    key_iter = iter(dict_iters)
    first_key = next(key_iter)
    first_iter = dict_iters[first_key]
    secondary_iters = {k: dict_iters[k] for k in key_iter}
    values = {}

    async def consume_iter(key, iter):
        # logger.debug(f"Starting consumer for {key}")
        async for item in iter:
            # logger.debug(f"Consumed {item} for {key}")
            values[key] = item

    tasks = [
        asyncio.create_task(consume_iter(k, iter))
        for k, iter in secondary_iters.items()
    ]
    async for item in first_iter:
        # logger.debug(f"Merging dict for {item}")
        yield {first_key: item, **values}
    for task in tasks:
        task.cancel()


def partition_step(
    async_iter: AsyncIterator[T], condition: Callable[[T], bool]
) -> Tuple[AsyncIterator[T], AsyncIterator[T]]:
    """Partitions a source into two iterators, one where the condition is true and one where it is false.
    Exceptions are propagated to both iterators."""
    true_queue = QueueWithException()
    false_queue = QueueWithException()

    async def distribute():
        try:
            async for item in async_iter:
                # logger.debug(f"Distributing {item}")
                if condition(item):
                    await true_queue.put(item)
                else:
                    await false_queue.put(item)
                # logger.debug(f"Queue sizes: {true_queue.qsize()} {false_queue.qsize()}")
        except Exception as e:
            # logger.debug(f"Exception while queueing")
            true_queue.set_exception(e)
            false_queue.set_exception(e)
        await true_queue.put(None)  # Signal end of iteration
        await false_queue.put(None)

    distribution_task = asyncio.create_task(distribute())
    true_iterator = _make_queue_iterator(true_queue)
    false_iterator = _make_queue_iterator(false_queue)

    return true_iterator, false_iterator


def fork_step(
    async_iter: AsyncIterator[T],
    backpressure: bool = True,
) -> Tuple[AsyncIterator[T], AsyncIterator[T]]:
    """Forks a stream, splitting a single source into two separate sources.
    If backpressure is True (the default), then items will only be consumed as fast as the first returned source is consumed,
    the second source will be rate limited to the first.  If false, items will always be consumed as fast as possible.

    If backpressure is True, then exceptions will only be propagated to the first iterator.  If False, it will go to both.
    """

    def with_backpressure():
        right_queue = asyncio.Queue()
        right_iterator = _make_queue_iterator(right_queue)

        async def consume_and_queue() -> AsyncIterator[T]:
            async for item in async_iter:
                await right_queue.put(item)
                yield item
            await right_queue.put(None)

        left_iterator = consume_and_queue()
        return left_iterator, right_iterator

    def no_backpressure():
        left_queue = QueueWithException()
        right_queue = QueueWithException()

        async def distribute():
            try:
                async for item in async_iter:
                    await left_queue.put(item)
                    await right_queue.put(item)
            except Exception as e:
                left_queue.set_exception(e)
                right_queue.set_exception(e)
            await left_queue.put(None)  # Signal end of iteration
            await right_queue.put(None)

        asyncio.create_task(distribute())
        left_iterator = _make_queue_iterator(left_queue)
        right_iterator = _make_queue_iterator(right_queue)

        return left_iterator, right_iterator

    iterators = with_backpressure() if backpressure else no_backpressure()
    return iterators


def buffer_step(
    async_iter: AsyncIterator[T], join_func: Callable[[T], T]
) -> AsyncIterator[T]:
    """Acts as a buffer that can receive multiple byte messages and aggregates them on the way out."""
    queue = QueueWithException()

    async def make_iterator() -> AsyncIterator[T]:
        ongoing = True
        while ongoing:
            # Use a blocking get() to ensure there's at least one chunk of
            # data, and stop iteration if the chunk is None, indicating the
            # end of the audio stream.
            try:
                item = await queue.get()
                if item is None:
                    # logger.info(f"None was on top of queue")
                    return
                data = [item]
            except asyncio.CancelledError:
                # logger.info(f"ByteBuffer cancelled")
                return
            # Now consume whatever other data's still buffered.
            while True:
                try:
                    chunk = queue.get_nowait()
                    if chunk is None:
                        # logger.info(f"Got None in nowait")
                        ongoing = False
                        break
                    data.append(chunk)
                except QueueEmpty:
                    # logger.info(f"Got QueueEmpty in nowait")
                    break
            yield join_func(data)

    asyncio.create_task(queue.enqueue_iterator(async_iter))
    return make_iterator()


def byte_buffer_step(async_iter: AsyncIterator[bytes]) -> AsyncIterator[bytes]:
    return buffer_step(async_iter, lambda x: b"".join(x))


def str_buffer_step(async_iter: AsyncIterator[str]) -> AsyncIterator[str]:
    return buffer_step(async_iter, lambda x: "".join(x))


async def _make_queue_iterator(
    queue: Union[asyncio.Queue, QueueWithException]
) -> AsyncIterator[T]:
    """Returns an async iterator over objects in a queue.  The queue must be closed by putting None into it."""
    while True:
        try:
            # logger.debug(f"Dequeuing {queue}")
            item = await queue.get()
            # logger.debug(f"Dequeued {item}")
            if item is None:
                break
        except asyncio.CancelledError:
            # logger.debug("Queue iterator cancelled.")
            break
        yield item



import asyncio
import datetime
import logging
from typing import AsyncIterator, Callable, Tuple

from voice_stream.basic_streams import log_step
from voice_stream.events import SpeechStart, BaseEvent, SpeechEnd

logger = logging.getLogger(__name__)

SpeechStep = Callable[
    [AsyncIterator[bytes]], Tuple[AsyncIterator[str], AsyncIterator[BaseEvent]]
]


def speech_with_start_detection_step(
    async_iter: AsyncIterator[bytes], speech_step: SpeechStep
):
    pipe, speech_events = speech_step(async_iter)
    # speech_events = log_step(speech_events, "Speech event")
    speech_start = filter_spurious_speech_start_events_step(speech_events)
    return pipe, speech_start


def filter_spurious_speech_start_events_step(
    async_iter: AsyncIterator[BaseEvent], threshold: float = 1.0
):
    """Detects the start of speech and returns the audio that was detected."""

    class WaitingIterator:
        def __init__(self):
            self.aiter = async_iter.__aiter__()
            self.task_for_next_item = None

        def __aiter__(self):
            return self

        async def __anext__(self):
            end_time = None
            start_event = None
            while True:
                now = datetime.datetime.now()
                timeout_to_use = (
                    (end_time - now).microseconds / 1e6 if end_time else None
                )
                if timeout_to_use and timeout_to_use <= 0:
                    timeout_to_use = None
                if not self.task_for_next_item:
                    self.task_for_next_item = asyncio.create_task(
                        self.aiter.__anext__()
                    )
                # logger.debug(f"Starting to wait with timeout {timeout_to_use}.")
                done, pending = await asyncio.wait(
                    {self.task_for_next_item}, timeout=timeout_to_use
                )
                if pending:
                    # We didn't get an end event, so this was a real speech start.
                    # logger.debug("No response received in time.  Sending start event.")
                    assert not done
                    assert start_event
                    return start_event
                else:
                    assert done
                    self.task_for_next_item = None
                    item = await done.pop()
                    if isinstance(item, SpeechStart):
                        # logger.debug("Received Speech Start.  Resetting timer")
                        end_time = datetime.datetime.now() + datetime.timedelta(
                            seconds=threshold
                        )
                        start_event = item
                    elif isinstance(item, SpeechEnd):
                        # logger.debug("Received Speech End, ignoring the start.")
                        end_time = None
                        start_event = None
                    else:
                        logger.info(f"Received unknown speech event, ignoring. {item}")

    pipe = WaitingIterator()
    pipe = log_step(pipe, "New Speech Detected", lambda x: x.time_since_start)
    return pipe



from __future__ import annotations

import asyncio
import base64
import dataclasses
import logging
from typing import AsyncIterator, Tuple, Optional

from quart import websocket, current_app

from voice_stream.basic_streams import (
    FutureOrObj,
    partition_step,
    resolve_obj_or_future,
    map_step,
    map_str_to_json_step,
    filter_step,
    fork_step,
    extract_value_step,
    map_future,
    merge_step,
    queue_source,
)
from voice_stream.events import BaseEvent, CallStarted, CallEnded
from voice_stream.integrations.quart_streams import quart_websocket_source

logger = logging.getLogger(__name__)


def twilio_split_media_step(
    async_iter: AsyncIterator[dict],
) -> Tuple[AsyncIterator[dict], AsyncIterator[dict]]:
    """Returns 2 iterators for a twilio stream, the media stream, and control messages."""
    return partition_step(async_iter, lambda x: x["event"] == "media")


async def twilio_check_sequence_step(
    async_iter: AsyncIterator[dict],
) -> AsyncIterator[dict]:
    """Verifies that no websocket messages were lost in the sequence"""
    expected = 1
    async for item in async_iter:
        assert "sequenceNumber" in item, f"Message {item} has no sequence number"
        if int(item["sequenceNumber"]) != expected:
            raise ValueError(
                f"Expected sequence number {expected}, got {item['sequenceNumber']}"
            )
        expected += 1
        yield item


async def twilio_close_on_stop_step(
    async_iter: AsyncIterator[dict],
) -> AsyncIterator[dict]:
    """Performs a websocket.close() when a stop message is received."""
    async for item in async_iter:
        if item["event"] == "stop":
            logger.info("Stop message received, closing websocket")
            await websocket.close(1000)
        yield item


async def twilio_media_to_audio_bytes_step(
    async_iter: AsyncIterator[dict],
) -> AsyncIterator[bytes]:
    """Converts a twilio media stream into a stream of audio bytes"""
    async for item in async_iter:
        assert "media" in item, f"Message {item} has no media"
        yield base64.b64decode(item["media"]["payload"])


async def audio_bytes_to_twilio_media_step(
    async_iter: AsyncIterator[bytes], stream_sid: FutureOrObj[str]
) -> AsyncIterator[dict]:
    resolved_stream_sid = None
    async for packet in async_iter:
        encoded_data = base64.b64encode(packet).decode("utf-8")
        if not resolved_stream_sid:
            resolved_stream_sid = await resolve_obj_or_future(stream_sid)
        yield {
            "event": "media",
            "streamSid": resolved_stream_sid,
            "media": {"payload": encoded_data},
        }


async def twilio_format_events_step(
    async_iter: AsyncIterator[bytes],
) -> AsyncIterator[BaseEvent]:
    def format_events(input):
        if input["event"] == "start":
            return CallStarted(
                call_id=input["start"]["callSid"], stream_id=input["start"]["streamSid"]
            )
        if input["event"] == "stop":
            return CallEnded()

    return map_step(async_iter, format_events)


@dataclasses.dataclass
class TwilioInputFlow:
    # Stream of audio bytes.
    audio: AsyncIterator[bytes]
    # All the events coming from the main stream in twilio, excluding raw audio.
    events: AsyncIterator[BaseEvent]
    all_twilio_messages: Optional[AsyncIterator[dict]]
    call_sid_f: asyncio.Future[str]
    stream_sid_f: asyncio.Future[str]
    inbound_queue_f: asyncio.Future[asyncio.Queue[BaseEvent]]
    outbound_queue_f: asyncio.Future[asyncio.Queue[BaseEvent]]

    @classmethod
    async def create(cls, expose_all_messages: bool = False) -> TwilioInputFlow:
        """
        Creates a basic flow to receive audio and call event data from Twilio.
        expose_all_messages - If true, all incoming twilio messages, including raw audio, will be set through the 'all_twilio_messages' iterator.  Otherwise this iterator will be empty.
        """
        pipe = quart_websocket_source()
        pipe = map_str_to_json_step(pipe)
        pipe = filter_step(pipe, lambda x: x["event"] != "connected")
        if expose_all_messages:
            pipe, all_twilio_messages = fork_step(pipe)
        else:
            all_twilio_messages = None
        pipe = twilio_check_sequence_step(pipe)
        # Retrieve values once the call is connected.
        pipe, call_sid_f = extract_value_step(
            pipe, value=lambda x: x["start"]["callSid"]
        )
        pipe, stream_sid_f = extract_value_step(pipe, value=lambda x: x["streamSid"])
        outbound_queue_f = map_future(
            call_sid_f, lambda x: current_app.current_calls[x].outbound
        )
        inbound_queue_f = map_future(
            call_sid_f, lambda x: current_app.current_calls[x].inbound
        )
        audio, events = partition_step(pipe, lambda x: x["event"] == "media")
        audio = twilio_media_to_audio_bytes_step(audio)
        events = twilio_close_on_stop_step(events)
        events = twilio_format_events_step(events)
        events = merge_step(events, queue_source(inbound_queue_f))
        return cls(
            audio,
            events,
            all_twilio_messages,
            call_sid_f,
            stream_sid_f,
            inbound_queue_f,
            outbound_queue_f,
        )



from voice_stream.audio.audio_streams import (
    ogg_page_separator_step,
    ogg_concatenator_step,
    wav_mulaw_file_source,
    wav_mulaw_file_sink,
)

from voice_stream.basic_streams import (
    array_sink,
    array_source,
    async_init_step,
    binary_file_sink,
    binary_file_source,
    concat_step,
    exception_handler_step,
    filter_step,
    fork_step,
    log_step,
    map_step,
    extract_value_step,
    merge_as_dict_step,
    merge_step,
    partition_step,
    queue_sink,
    queue_source,
    str_buffer_step,
    text_file_sink,
    text_file_source,
    count_step,
    byte_buffer_step,
    chunk_bytes_step,
    flatten_step,
    map_str_to_json_step,
)
from voice_stream.speech_to_text_streams import (
    filter_spurious_speech_start_events_step,
    speech_with_start_detection_step,
)

from voice_stream.substreams import (
    substream_on_dict_key_step,
    substream_step,
    interruptable_substream_step,
    cancelable_substream_step,
)

from voice_stream.text_to_speech_streams import (
    raw_audio_rate_limit_step,
    timed_text_rate_limit_step,
    audio_rate_limit_step,
    buffer_tts_text_step,
    tts_rate_limit_step,
    tts_with_buffer_and_rate_limit_step,
    wait_for_punctuation_step,
    min_tokens_step,
    AudioWithText,
)

__all__ = [
    array_sink,
    array_source,
    async_init_step,
    audio_rate_limit_step,
    AudioWithText,
    binary_file_sink,
    binary_file_source,
    buffer_tts_text_step,
    byte_buffer_step,
    cancelable_substream_step,
    chunk_bytes_step,
    concat_step,
    count_step,
    exception_handler_step,
    extract_value_step,
    filter_step,
    filter_spurious_speech_start_events_step,
    flatten_step,
    fork_step,
    interruptable_substream_step,
    log_step,
    map_step,
    map_str_to_json_step,
    merge_as_dict_step,
    merge_step,
    min_tokens_step,
    partition_step,
    queue_sink,
    queue_source,
    raw_audio_rate_limit_step,
    speech_with_start_detection_step,
    substream_on_dict_key_step,
    substream_step,
    text_file_sink,
    text_file_source,
    timed_text_rate_limit_step,
    tts_rate_limit_step,
    tts_with_buffer_and_rate_limit_step,
    wait_for_punctuation_step,
]



import asyncio
import logging
from typing import (
    AsyncIterator,
    Callable,
    Tuple,
    TypeVar,
    Any,
    Optional,
    List,
    Union,
)

from voice_stream._substream_iters import SwitchableIterator
from voice_stream.basic_streams import (
    single_source,
    queue_source,
    empty_source,
    concat_step,
    to_tuple,
    from_tuple,
    empty_sink,
)

T = TypeVar("T")
Output = TypeVar("Output")

logger = logging.getLogger(__name__)


async def substream_on_dict_key_step(
    async_iter: AsyncIterator[dict], key: str, substream_func
) -> AsyncIterator[dict]:
    current_dict = None

    async def input_gen():
        async for item in async_iter:
            nonlocal current_dict
            current_dict = item
            yield item[key]

    pipe = input_gen()
    pipe = substream_func(pipe)
    async for item in pipe:
        ret = {**current_dict, key: item}
        yield ret


async def substream_step(async_iter: AsyncIterator[T], substream_func):
    """For each item in the input iterator, creates a substream and feeds the item into it."""
    async for item in async_iter:
        pipe = single_source(item)
        pipe = substream_func(pipe)
        async for sub_item in pipe:
            yield sub_item


def cancelable_substream_step(
    async_iter: AsyncIterator[T],
    cancel_iter: AsyncIterator[T],
    substream_func: Callable[[AsyncIterator[T]], Union[AsyncIterator[Output], Tuple]],
    cancel_messages: Optional[List[Any]] = None,
):
    """Creates a new substream for each input to async_iter.  If any item comes in on cancel_iter, it immediately stops
    the processing of the current substream, and optionally sends cancel_messages.
    Returns as many AsyncIterators as are returned by the substream func.
    """

    substream_completed = asyncio.Event()

    async def monitor_cancel():
        nonlocal next_substream, completed_outputs
        async for _ in cancel_iter:
            try:
                if completed_outputs < len(output_iters):
                    for i in output_iters:
                        i.disconnect()
                    cancel_iters = _create_cancel_messages(
                        output_iters, cancel_messages
                    )
                    completed_outputs = 0
                    for o, c in zip(output_iters, cancel_iters):
                        o.switch(c)
                else:
                    logger.debug(
                        "Ignoring cancel because there is no active substream."
                    )
            except Exception as e:
                logger.error(f"Error in cancelable_substream: {e}")

    def on_output_complete():
        # logger.debug("on_output_complete")
        nonlocal completed_outputs
        completed_outputs += 1
        if completed_outputs == len(output_iters):
            substream_completed.set()

    next_source = queue_source()
    next_substream = to_tuple(substream_func(next_source))
    output_iters = [
        SwitchableIterator(
            None, callback=on_output_complete, propagate_end_of_iter=False
        )
        for _ in next_substream
    ]
    completed_outputs = len(output_iters)
    cancel_messages = (
        cancel_messages if cancel_messages else [None for _ in output_iters]
    )

    monitor_cancel_task = asyncio.create_task(monitor_cancel())

    async def gen():
        nonlocal next_source, next_substream, completed_outputs
        async for item in async_iter:
            completed_outputs = 0
            substream_completed.clear()
            _switch_outputs(output_iters, next_substream)
            await next_source.put(item)
            await next_source.put(None)
            await substream_completed.wait()
            next_source = queue_source()
            next_substream = to_tuple(substream_func(next_source))
        monitor_cancel_task.cancel()
        await next_source.put(None)
        await asyncio.wait([empty_sink(i) for i in next_substream])
        for i in output_iters:
            i.end_iteration()

    asyncio.create_task(gen())
    return from_tuple(output_iters)


def interruptable_substream_step(
    async_iterator: AsyncIterator[T],
    substream_func: Callable[
        [AsyncIterator[T]],
        Callable[[AsyncIterator[T]], Union[AsyncIterator[Output], Tuple]],
    ],
    cancel_messages: Optional[List[Any]] = None,
):
    """For each input, creates a new substream and runs it.
    If a new item comes in, the old item is immediately cancelled and a new one is created.
    """

    completed_outputs = 0

    def on_output_complete():
        nonlocal completed_outputs
        completed_outputs += 1

    active_source = None
    next_source = queue_source()
    next_substream = to_tuple(substream_func(next_source))
    # logger.debug(f"Created substream {next_substream}")
    output_iters = [
        SwitchableIterator(_, callback=on_output_complete, propagate_end_of_iter=False)
        for _ in next_substream
    ]

    async def gen():
        nonlocal active_source, next_source, next_substream, completed_outputs
        async for item in async_iterator:
            if active_source:
                if cancel_messages and completed_outputs < len(output_iters):
                    cancel_iters = _create_cancel_messages(
                        output_iters, cancel_messages
                    )
                    next_substream = [
                        concat_step(cancel_iter, output)
                        for output, cancel_iter in zip(next_substream, cancel_iters)
                    ]
                completed_outputs = 0
                _switch_outputs(output_iters, next_substream)
            # logger.debug("Running substream")
            active_source = next_source
            await active_source.put(item)
            await active_source.put(None)
            next_source = queue_source()
            next_substream = to_tuple(substream_func(next_source))
        logger.debug("Completed interruptable substream iter")
        for i in output_iters:
            i.end_iteration()

    logger.debug("Creating task")
    asyncio.create_task(gen())
    return from_tuple(output_iters)


def _create_cancel_messages(outputs, cancel_messages):
    if cancel_messages and len(cancel_messages) != len(outputs):
        raise ValueError(
            f"cancel_messages must be the same length as the number of outputs for the substream.  Got {len(cancel_messages)} cancel messages and {len(output_iters)} substream outputs."
        )

    def create_cancel(x):
        if callable(x):
            return x()
        elif x:
            logger.debug(f"SINGLE SOURCE {x}")
            return single_source(x)
        else:
            return empty_source()

    return [create_cancel(_) for _ in cancel_messages]


def _switch_outputs(
    switchable_iters: List[SwitchableIterator], new_outputs: List[AsyncIterator]
):
    assert len(new_outputs) == len(switchable_iters)
    for switchable_iter, new_output in zip(switchable_iters, new_outputs):
        switchable_iter.switch(new_output)



import asyncio
import inspect
import logging
from asyncio import InvalidStateError
from typing import AsyncIterator, Any, TypeVar

T = TypeVar("T")

logger = logging.getLogger(__name__)


class SimpleFlow(AsyncIterator):
    """Serves as a stub for creating flows.  Acts like an async iterator, but when apply is called with an async iterator, it wraps that iterator."""

    def __init__(self):
        loop = asyncio.get_running_loop()
        self.iter_f = loop.create_future()
        self.iter = None

    def finish(self, *outputs: AsyncIterator[Any]):
        self.outputs = outputs

    def __call__(self, async_iter: AsyncIterator[T]):
        if not self.outputs:
            raise InvalidStateError("Flow was not finished with finish()")
        self.iter_f.set_result(async_iter)
        return self.outputs

    def flow_for_output(self, ix: int):
        def output_wrapper(async_iter: AsyncIterator[T]):
            ret = self.__call__(async_iter)
            return ret[ix]

        return output_wrapper

    def __aiter__(self):
        if self.iter:
            self.iter = self.iter_f.result().__aiter__()
        return self

    async def __anext__(self):
        if not self.iter:
            logger.info(
                "Stub iterator not resolved before use.  If the stream stalls, this could be the problem."
            )
            self.iter = (await self.iter_f).__aiter__()
        return await self.iter.__anext__()


class ResettableIterator(AsyncIterator):
    """Wraps an AsyncIterator and allows it to be used in multiple downstream generators.  It works like a normal
    iterator, but if 'reset' is called, it throws a StopAsyncIteration.  This ends iteration for all downstream tasks.
     However, the underlying iterator isn't finished and so this iterator can be used in another generator.
    """

    def __init__(self, async_iter: AsyncIterator):
        self.async_iter = async_iter
        self.reset_flag = asyncio.Event()
        self.next_item_task = None

    def __aiter__(self):
        return self

    async def __anext__(self):
        # logger.trace(f"Entering __anext__")
        if self.reset_flag.is_set():
            # logger.trace(f"Reset flag set on entry")
            raise StopAsyncIteration
        if self.next_item_task is None:
            self.next_item_task = asyncio.create_task(self.async_iter.__anext__())
        reset_task = asyncio.create_task(self.reset_flag.wait())
        done, pending = await asyncio.wait(
            {self.next_item_task, reset_task}, return_when=asyncio.FIRST_COMPLETED
        )
        # logger.trace(f"Tasks completed: Done is {done}.  Pending is {pending}")
        if reset_task in done:
            # logger.debug(f"Reset flag set on wait")
            raise StopAsyncIteration
        assert reset_task in pending and self.next_item_task in done
        reset_task.cancel()
        ret = await self.next_item_task
        self.next_item_task = None
        return ret

    def reset(self):
        self.reset_flag.set()
        # It's important that we create a new event because we don't know if __anext__ will be called again before the original event is cleared.
        self.reset_flag = asyncio.Event()


class SwitchableIterator:
    """
    An iterator that can change its source.

    It is created with an iterator.  When disconnect is called, it cancels looking at the original iterator and waits for a call to switch.
    When switch is called, moves over to the pull from the new iterator.

    If None is passed for the initial async_iter, then nothing will happen until switch is called.

    The callback allows you to specify whether to propagate a StopAsyncIteration exception.  If you don't propagate, then you must eventually call end_iteration.

    The state model for the initial iterator

    Connected - The iterator is actively forwarding from a source iterator.  Any call to __anext__ will be forwarded to that iterator.
    Disconnected - The is not connected.  Any call to __anext__ will block until the iterator is connected.
    Completed - The iterator is done, it has sent an end of iteration message.
    """

    CONNECTED = "Connected"
    DISCONNECTED = "Disconnected"
    COMPLETED = "Completed"

    def __init__(
        self,
        async_iter: AsyncIterator,
        callback=None,
        propagate_end_of_iter: bool = True,
    ):
        self.state = (
            SwitchableIterator.CONNECTED
            if async_iter
            else SwitchableIterator.DISCONNECTED
        )
        self.state_change_event = asyncio.Event()
        self.async_iter = async_iter
        self.callback = callback
        self.propagate_end_of_iter = propagate_end_of_iter
        # logger.debug(f"New SwitchableIterator {self}")

    def __aiter__(self):
        return self

    async def __anext__(self):
        # logger.debug(f"Entering SwitchableIterator.__anext__ {self}")
        while True:
            self.state_change_event.clear()
            try:
                if self.state == SwitchableIterator.DISCONNECTED:
                    await self.state_change_event.wait()
                elif self.state == SwitchableIterator.COMPLETED:
                    raise StopAsyncIteration
                else:
                    assert self.state == SwitchableIterator.CONNECTED
                    next_item_task = asyncio.create_task(self.async_iter.__anext__())
                    state_change_task = asyncio.create_task(
                        self.state_change_event.wait()
                    )
                    done, pending = await asyncio.wait(
                        {next_item_task, state_change_task},
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    for task in pending:
                        task.cancel()
                        try:
                            await task
                        except asyncio.CancelledError:
                            pass
                        except StopAsyncIteration:
                            pass
                    if next_item_task in done:
                        try:
                            return await next_item_task
                        except StopAsyncIteration as e:
                            if self.callback:
                                self.callback()
                            if self.propagate_end_of_iter:
                                raise e
                            # If we aren't propagating the end, disconnect the iterator.
                            self.disconnect()
            finally:
                self.state_change_event.clear()

    def end_iteration(self):
        self.state = SwitchableIterator.COMPLETED
        self.async_iter = None
        self.state_change_event.set()

    def switch(self, async_iter: AsyncIterator):
        if not inspect.isasyncgen(async_iter):  # and not is_async_iterator(async_iter):
            raise ValueError(
                f"SwitchableIterator.switch must be called with an AsyncIterator.  Got {async_iter}"
            )
        # logger.debug("Switching iterators")
        if self.state == SwitchableIterator.COMPLETED:
            raise InvalidStateError("Cannot switch a completed iterator")
        self.state = SwitchableIterator.CONNECTED
        self.async_iter = async_iter
        self.state_change_event.set()

    def disconnect(self):
        # logger.debug("Disconnecting iterators")
        if self.state == SwitchableIterator.COMPLETED:
            raise InvalidStateError("Cannot disconnect a completed iterator")
        self.state = SwitchableIterator.DISCONNECTED
        self.async_iter = None
        self.state_change_event.set()



import dataclasses
import logging
from typing import AsyncIterator, Union

from google.api_core.exceptions import Aborted
from google.cloud.speech_v2 import (
    StreamingRecognizeRequest,
    StreamingRecognitionConfig,
    StreamingRecognitionFeatures,
    SpeechAsyncClient,
    StreamingRecognizeResponse,
    RecognitionConfig,
)
from google.cloud.texttospeech_v1 import (
    TextToSpeechAsyncClient,
    SynthesizeSpeechRequest,
    SynthesisInput,
    VoiceSelectionParams,
    AudioConfig,
    AudioEncoding,
)

from voice_stream.audio.audio_streams import remove_wav_header, AudioFormat
from voice_stream.basic_streams import (
    map_step,
    filter_step,
    array_source,
    concat_step,
    byte_buffer_step,
    chunk_bytes_step,
    exception_handler_step,
    partition_step,
    async_init_step,
)
from voice_stream.events import SpeechStart, SpeechEnd
from voice_stream.integrations.google_setup import (
    resolve_audio_decoding,
    GoogleDecodingConfig,
)
from voice_stream.text_to_speech_streams import AudioWithText

# Can be one of the standard audio formats or a Google AudioConfig object
GoogleAudioConfig = Union[AudioFormat, AudioConfig]

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class TTSRequest:
    text: str
    voice: str


def google_speech_step(
    async_iter: AsyncIterator[bytes],
    speech_async_client: SpeechAsyncClient,
    project,
    location,
    recognizer,
    model: str = "latest_long",
    language_codes: Union[str, list[str]] = ["en-US", "es-US"],
    audio_format: GoogleDecodingConfig = None,
    include_events: bool = False,
) -> AsyncIterator[str]:
    """Converts a stream of audio bytes into a stream of text.
    If include_events is True, a second stream will be returned containing speech events.
    """
    MAX_STREAM_SIZE = 25600
    pipe = async_iter
    # pipe = count_step(pipe, "pre")
    pipe = byte_buffer_step(pipe)
    pipe = chunk_bytes_step(pipe, MAX_STREAM_SIZE)
    # pipe = count_step(pipe, "post")
    pipe = map_step(
        pipe, lambda x: StreamingRecognizeRequest(recognizer=recognizer, audio=x)
    )
    config = array_source(
        [
            _initial_recognition_config(
                include_events,
                project,
                location,
                recognizer,
                model,
                language_codes,
                audio_format,
            )
        ]
    )
    pipe = concat_step(config, pipe)
    pipe = async_init_step(pipe, speech_async_client.streaming_recognize)
    pipe = exception_handler_step(
        pipe, Aborted, lambda x: logger.info("Google Recognize aborted.")
    )
    # pipe = filter_step(pipe, lambda x: x.results[0].is_final)
    # pipe = log_step(pipe, "google_speech")

    def get_transcript(result):
        if not result.results:
            logger.info("No results in google response")
            return ""
        if not result.results[0].alternatives:
            logger.info("No alternatives in google response")
            return ""
        return result.results[0].alternatives[0].transcript

    if include_events:
        pipe, events = partition_step(pipe, lambda x: x.results)
        # pipe = log_step(pipe, "google_partition")
    pipe = map_step(pipe, get_transcript)
    # pipe = log_step(pipe, "get_transcript")
    pipe = filter_step(pipe, lambda x: len(x) > 0)
    if include_events:
        events = map_step(events, _map_speech_events)
        events = filter_step(events, lambda x: x is not None)
        # pipe = log_step(pipe, "Google out")
        return pipe, events
    else:
        return pipe


async def google_text_to_speech_step(
    async_iter: AsyncIterator[Union[str, TTSRequest]],
    text_to_speech_async_client: TextToSpeechAsyncClient,
    voice_name: str = "en-US-Standard-H",
    audio_format: GoogleAudioConfig = AudioFormat.OGG_OPUS,
) -> AsyncIterator[AudioWithText]:
    """Each text block that comes in is converted to audio using the desired voice and format.
    Can take either strings in, or a TTSRequest, which allows you to specify the voice.
    """
    audio_config = _resolve_google_audio_config(audio_format)

    async for item in async_iter:
        if isinstance(item, TTSRequest):
            voice = item.voice
            text = item.text
        else:
            voice = voice_name
            text = item
        # noinspection PyTypeChecker
        request = SynthesizeSpeechRequest(
            input=SynthesisInput(text=text),
            voice=VoiceSelectionParams(
                language_code="en-us",
                name=voice,
            ),
            audio_config=audio_config,
        )
        result = await text_to_speech_async_client.synthesize_speech(request)
        audio = result.audio_content
        if audio_config.audio_encoding == AudioEncoding.MULAW:
            audio = remove_wav_header(audio)
        yield AudioWithText(audio=audio, text=text, audio_format=audio_format)


def _resolve_google_audio_config(audio_format: GoogleAudioConfig):
    if isinstance(audio_format, AudioConfig):
        return audio_format
    elif audio_format == AudioFormat.WAV_MULAW_8KHZ:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.MULAW,
            sample_rate_hertz=8000,
        )
    elif audio_format == AudioFormat.OGG_OPUS:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.OGG_OPUS,
        )
    elif audio_format == AudioFormat.MP3:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.MP3,
        )
    else:
        assert False, f"Unsupported audio format {audio_format}"


def _initial_recognition_config(
    include_events: bool,
    project: str,
    location: str,
    recognizer: str,
    model: str = "latest_long",
    language_codes: Union[str, list[str]] = ["en-US", "es-US"],
    audio_format: GoogleDecodingConfig = None,
):
    recognizer_path = (
        f"projects/{project}/locations/{location}/recognizers/{recognizer}"
    )
    auto_decoding, explicit_decoding = resolve_audio_decoding(audio_format)

    # noinspection PyTypeChecker
    ret = StreamingRecognizeRequest(
        recognizer=recognizer_path,
        streaming_config=StreamingRecognitionConfig(
            config=RecognitionConfig(
                auto_decoding_config=auto_decoding,
                explicit_decoding_config=explicit_decoding,
                model=model,
                language_codes=language_codes,
            ),
            streaming_features=StreamingRecognitionFeatures(
                interim_results=False,
                enable_voice_activity_events=include_events,
                # voice_activity_timeout=StreamingRecognitionFeatures.VoiceActivityTimeout(
                #     # speech_end_timeout
                # ),
            ),
        ),
    )
    # logger.debug(f"Initial recognition config: {ret}")
    return ret


def _event_time_to_seconds(event):
    return event.speech_event_offset.seconds + event.speech_event_offset.nanos / 1e9


def _map_speech_events(input):
    if (
        input.speech_event_type
        == StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_BEGIN
    ):
        return SpeechStart(time_since_start=input.speech_event_offset.total_seconds())
    if (
        input.speech_event_type
        == StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_END
    ):
        return SpeechEnd(time_since_start=input.speech_event_offset.total_seconds())
    return None



import json
import logging
from asyncio import CancelledError
from typing import AsyncIterator, Union
from fastapi import WebSocket
from hypercorn.utils import UnexpectedMessageError
from starlette.websockets import WebSocketDisconnect

logger = logging.getLogger(__name__)


async def fastapi_websocket_text_source(websocket: WebSocket) -> AsyncIterator[str]:
    """Creates an AsyncIterator for text messages coming from the FastAPI websocket."""
    await websocket.accept()
    try:
        while True:
            message = await websocket.receive_text()
            yield message
    except WebSocketDisconnect:
        logger.debug("websocket.receive is cancelled.")


async def fastapi_websocket_bytes_source(websocket: WebSocket) -> AsyncIterator[str]:
    """Creates an AsyncIterator for binary data coming from the FastAPI websocket."""
    await websocket.accept()
    try:
        while True:
            message = await websocket.receive_bytes()
            yield message
    except WebSocketDisconnect:
        logger.debug("websocket.receive is cancelled.")


async def fastapi_websocket_text_sink(
    async_iter: AsyncIterator[Union[str, dict]], websocket: WebSocket
) -> None:
    """Takes an async iterator and sends everything to a FastAPI websocket.  Handles strings or dicts."""
    async for message in async_iter:
        if isinstance(message, dict):
            message = json.dumps(message)
        await websocket.send_text(message)


async def fastapi_websocket_binary_sink(
    async_iter: AsyncIterator[bytes], websocket: WebSocket
) -> None:
    """Takes an async iterator and sends everything to a FastAPI websocket.  Expects bytes."""
    async for message in async_iter:
        await websocket.send_bytes(message)



import dataclasses


class BaseEvent:
    event_name: str


@dataclasses.dataclass
class SpeechStart(BaseEvent):
    time_since_start: float
    event_name: str = "speech_start"


@dataclasses.dataclass
class SpeechEnd(BaseEvent):
    time_since_start: float
    event_name: str = "speech_end"


@dataclasses.dataclass
class TextInput(BaseEvent):
    text: str
    event_name: str = "text_input"


@dataclasses.dataclass
class TextOutput(BaseEvent):
    text: str
    event_name: str = "text_output"


@dataclasses.dataclass
class TimedText(BaseEvent):
    text: str
    duration_in_seconds: float
    event_name: str = "timed_text"


@dataclasses.dataclass
class AnsweringMachineDetection(BaseEvent):
    call_id: str
    answered_by: str
    time_since_start: float
    event_name: str = "amd"

    def is_human(self):
        return self.answered_by == "human"


@dataclasses.dataclass
class CallStarted(BaseEvent):
    call_id: str
    stream_id: str
    event_name: str = "call_started"


@dataclasses.dataclass
class CallEnded(BaseEvent):
    event_name: str = "call_ended"



import logging
from typing import Union, Optional

from google.cloud.speech_v2 import (
    Recognizer,
    CreateRecognizerRequest,
    SpeechAsyncClient,
    RecognitionConfig,
    ExplicitDecodingConfig,
    AutoDetectDecodingConfig,
)

from voice_stream.audio.audio_streams import AudioFormat

GoogleDecodingConfig = Optional[Union[AudioFormat, ExplicitDecodingConfig]]

logger = logging.getLogger(__name__)


async def create_recognizer(
    client: SpeechAsyncClient,
    project_id: str,
    location: str,
    recognizer_id: str,
    model: str,
    language_codes: Union[str, list[str]],
    audio_format: GoogleDecodingConfig = None,
):
    # Create a recognizer configuration
    parent = f"projects/{project_id}/locations/{location}"
    auto_decoding, explicit_decoding = resolve_audio_decoding(audio_format)

    # Create the recognizer
    # noinspection PyTypeChecker
    operation = client.create_recognizer(
        request=CreateRecognizerRequest(
            parent=parent,
            recognizer=Recognizer(
                name=f"{parent}/recognizers/{recognizer_id}",
                default_recognition_config=RecognitionConfig(
                    auto_decoding_config=auto_decoding,
                    explicit_decoding_config=explicit_decoding,
                    model=model,
                    language_codes=language_codes,
                ),
            ),
            recognizer_id=recognizer_id,
        )
    )

    # Wait for the operation to complete
    response = (await operation).result(timeout=90)
    response = await response
    logger.info(f"Recognizer created: {response.name}")


def resolve_audio_decoding(audio_format):
    if isinstance(audio_format, ExplicitDecodingConfig):
        return None, audio_format
    elif audio_format is None or audio_format != AudioFormat.WAV_MULAW_8KHZ:
        return AutoDetectDecodingConfig(), None
    else:
        # noinspection PyTypeChecker
        explicit = ExplicitDecodingConfig(
            encoding=ExplicitDecodingConfig.AudioEncoding.MULAW,
            sample_rate_hertz=8000,
            audio_channel_count=1,
        )
        return None, explicit



import asyncio
import logging
import struct
from concurrent.futures import ThreadPoolExecutor

import aiofiles

logger = logging.getLogger(__name__)


class AsyncMuLawStreamReader:
    def __init__(self, filename, chunk_size=4096):
        self.filename = filename
        self.chunk_size = chunk_size
        self._file = None

    async def open(self):
        self._file = await aiofiles.open(self.filename, "rb")
        # RIFF header
        riff_chunk_id = await self._file.read(4)
        if riff_chunk_id != b"RIFF":
            raise Exception("No RIFF header found")
        riff_chunk_size = struct.unpack("<I", await self._file.read(4))[0]

        wave_format = await self._file.read(4)
        if wave_format != b"WAVE":
            raise Exception("Not a WAVE file")

        # read fmt chunk
        fmt_chunk_id = await self._file.read(4)
        if fmt_chunk_id != b"fmt ":
            raise Exception("fmt chunk missing")

        fmt_chunk_size = struct.unpack("<I", await self._file.read(4))[0]
        self._audioformat = struct.unpack("<H", await self._file.read(2))[0]
        self._numchannels = struct.unpack("<H", await self._file.read(2))[0]
        self._samplerate = struct.unpack("<I", await self._file.read(4))[0]
        self._byterate = struct.unpack("<I", await self._file.read(4))[0]
        self._bytespersample = struct.unpack("<H", await self._file.read(2))[0]
        self._bitspersample = struct.unpack("<H", await self._file.read(2))[0]
        if fmt_chunk_size == 18:
            self._extraparams = struct.unpack("<H", await self._file.read(2))[0]
            fact_chunk_id = await self._file.read(4)
            if fact_chunk_id != b"fact":
                raise Exception("fact chunk missing")

            fact_chunk_size = struct.unpack("<I", await self._file.read(4))[0]
            self._samplelength = struct.unpack("<I", await self._file.read(4))[0]
        data_chunk_id = await self._file.read(4)
        if data_chunk_id != b"data":
            raise Exception("data chunk missing")
        data_chunk_size = struct.unpack("<I", await self._file.read(4))[0]
        if self._samplelength == 0:  # no fact chunk
            self._samplelength = data_chunk_size
        logger.info(f"Opened {self.filename} with {self._samplelength} samples")

    async def read(self):
        if not self._file:
            raise Exception("Wave file not opened. Call open() method first.")
        data = await self._file.read(self.chunk_size)
        return data

    async def close(self):
        if self._file:
            await self._file.close()


class AsyncMuLawStreamWriter:
    def __init__(
        self, filename, channels=1, sample_rate=8000, bits_per_sample=8, audio_format=7
    ):
        self.filename = filename
        self._numofchannels = channels
        self._samplerate = sample_rate
        self._bitspersample = bits_per_sample
        if audio_format != 1 and audio_format != 6 and audio_format != 7:
            raise Exception("Unsupported audio format")
        self._audioformat = audio_format
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.loop = asyncio.get_event_loop()
        self._file = None
        self._datawritten = 0

    async def open(self):
        self._file = await aiofiles.open(self.filename, "wb")
        await self._file.write(b"RIFF")
        self._form_length_pos = await self._file.tell()
        if self._audioformat == 1:
            await self._file.write(
                struct.pack(
                    "<L4s4sLHHLLHH4s",
                    36 + self._datawritten,
                    b"WAVE",
                    b"fmt ",
                    16,
                    self._audioformat,
                    self._numofchannels,
                    self._samplerate,
                    int(
                        self._numofchannels
                        * self._samplerate
                        * (self._bitspersample / 8)
                    ),
                    int(self._numofchannels * (self._bitspersample / 8)),
                    self._bitspersample,
                    b"data",
                )
            )
        elif self._audioformat == 6 or self._audioformat == 7:
            await self._file.write(
                struct.pack(
                    "<L4s4sLHHLLHHH4sLL4s",
                    50 + self._datawritten,
                    b"WAVE",
                    b"fmt ",
                    18,
                    self._audioformat,
                    self._numofchannels,
                    self._samplerate,
                    int(
                        self._numofchannels
                        * self._samplerate
                        * (self._bitspersample / 8)
                    ),
                    int(self._numofchannels * (self._bitspersample / 8)),
                    self._bitspersample,
                    0,
                    b"fact",
                    4,
                    self._datawritten,
                    b"data",
                )
            )
        self._data_length_pos = await self._file.tell()
        await self._file.write(struct.pack("<L", self._datawritten))

    async def write(self, audio_data):
        if not self._file:
            raise Exception("Wave file not opened. Call open() method first.")

        await self._file.write(audio_data)
        self._datawritten += len(audio_data)

    async def close(self):
        if self._file:
            await self._patchheader()
            await self._file.close()

    async def _patchheader(self):
        curpos = await self._file.tell()
        await self._file.seek(self._form_length_pos, 0)
        await self._file.write(struct.pack("<L", 36 + self._datawritten))
        await self._file.seek(self._data_length_pos, 0)
        await self._file.write(struct.pack("<L", self._datawritten))
        await self._file.seek(curpos, 0)



import asyncio
import io
import logging
from typing import AsyncIterator, Union

from mutagen.mp3 import MP3

from voice_stream.audio.async_wav_file import (
    AsyncMuLawStreamWriter,
    AsyncMuLawStreamReader,
)
from voice_stream.audio import AudioFormatError, AudioFormat
from voice_stream.audio.audio_mp3 import find_frame_boundaries, calculate_split_points
from voice_stream.audio.audio_ogg import OggPage, OpusIdPacket
from voice_stream.basic_streams import resolve_obj_or_future, FutureOrObj

logger = logging.getLogger(__name__)


async def wav_mulaw_file_source(
    filename: str, chunk_size: int = 4096
) -> AsyncIterator[bytes]:
    """Reads a wav file and returns a stream of audio bytes"""
    f = AsyncMuLawStreamReader(filename, chunk_size=chunk_size)
    try:
        await f.open()
        while True:
            data = await f.read()
            if not data:
                break
            # logger.debug(f"Read {len(data)} bytes from {filename}")
            yield data
    finally:
        await f.close()


async def wav_mulaw_file_sink(
    async_iter: AsyncIterator[bytes], filename: Union[str, asyncio.Future[str]]
) -> None:
    """Writes a stream of audio bytes to a wav file"""
    f = None
    try:
        async for message in async_iter:
            # We wait to resolve the filename until we have a message to write
            if f is None:
                filename = await resolve_obj_or_future(filename)
                f = AsyncMuLawStreamWriter(filename)
                await f.open()
            if message is None:
                break
            # logger.debug(f"Wrote {len(message)} bytes to {filename}")
            await f.write(message)
    finally:
        if f is not None:
            await f.close()
        logger.info(f"Closed {filename}")


async def mp3_chunk_step(
    async_iter: AsyncIterator[bytes], chunk_size: FutureOrObj[int]
) -> AsyncIterator[bytes]:
    """Takes in MP3 data and splits it on MP3 frame boundaries, making chunks as large as possible, but not larger than max_size, unless an individual frame is larger than max_size."""
    async for data in async_iter:
        resolved_chunk_size = await resolve_obj_or_future(chunk_size)
        boundaries = find_frame_boundaries(data)
        splits = calculate_split_points(boundaries, len(data), resolved_chunk_size)
        split_start = 0
        for split_end in splits:
            yield data[split_start:split_end]
            split_start = split_end
        assert split_end == len(data)


async def ogg_page_separator_step(
    async_iter: AsyncIterator[bytes],
) -> AsyncIterator[bytes]:
    """Takes in a stream of bytes from an OGG media file and outputs bytes ensuring that each output is a complete page.
    Checks if the last page in each chunk sent is a full page, if so, it sends it."""
    buffer = b""
    async for data in async_iter:
        concatenated = buffer + data
        if concatenated[:4] != b"OggS":
            raise AudioFormatError(
                f"Bytes didn't start with a proper OggS head.  Expected 'OggS, got {concatenated[:4]}"
            )
        position = concatenated.rfind(b"OggS")
        if OggPage.is_full_page(concatenated, position):
            position = len(concatenated)
        buffer = concatenated[position:]
        if position > 0:
            yield concatenated[:position]
    if len(buffer) > 0:
        yield buffer


async def ogg_concatenator_step(
    async_iter: AsyncIterator[bytes],
) -> AsyncIterator[bytes]:
    """With files in OGG format, you can't concatenate the bytes.  This performs the necessary operations to concatenate streams.
    This step assumes data is coming in chunked into full OGG pages.  If it is not, put an ogg_page_separator_step before.

    This could be done more efficiently by using numpy to modify the buffer in place."""
    sequence_offset = 0
    granule_offset = 0
    header = None

    async for data in async_iter:
        # Iterate through each ogg page and update.
        position = 0
        pages = []
        while position < len(data):
            page = OggPage.from_data(data, position)
            page_header_type = page.header_type
            position += page.page_length
            if page.page_sequence_number == 0:
                if not OpusIdPacket.is_opus_encoded(
                    data, page.offset + page.header_length
                ):
                    raise AudioFormatError(
                        "OGG file is not OPUS encoded.  Only Opus is currently supported."
                    )
                new_header = OpusIdPacket.from_data(
                    data, page.offset + page.header_length
                )
                if header is not None:
                    if (
                        header.output_channel_count != new_header.output_channel_count
                        or header.input_sample_rate != new_header.input_sample_rate
                        or header.output_gain != new_header.output_gain
                        or header.channel_mapping_family
                        != new_header.channel_mapping_family
                    ):
                        raise AudioFormatError(
                            "Cant concatenate OGG streams with different headers."
                        )
            if sequence_offset and page.page_sequence_number < 2:
                logger.debug("skipping opus header for later sequence")
                continue
            if sequence_offset:
                page = page.update(
                    new_header_type=0,
                    granule_offset=granule_offset,
                    sequence_offset=sequence_offset,
                )
                # page.log()
            if page_header_type == 4:
                sequence_offset = sequence_offset + page.page_sequence_number
                granule_offset = granule_offset + page.granule
            pages.append(page)
        logger.debug(f"Adjusted {len(pages)} OGG pages")
        adjusted = b"".join(_.get_bytes() for _ in pages)
        yield adjusted


def remove_wav_header(wav_bytes: bytes) -> bytes:
    """Removes the wav header from a wav file, regardless of the format."""
    # Check if it's a valid RIFF file
    if wav_bytes[:4] != b"RIFF" or wav_bytes[8:12] != b"WAVE":
        raise ValueError("Invalid WAV file")

    # Find the "data" chunk
    index = 12
    while index < len(wav_bytes):
        chunk_id = wav_bytes[index : index + 4]
        chunk_size = int.from_bytes(wav_bytes[index + 4 : index + 8], "little")
        if chunk_id == b"data":
            # Found the "data" chunk, return its content
            start = index + 8  # Skip "data" chunk header
            return wav_bytes[start : start + chunk_size]
        # logger.info(f"Skipping chunk {chunk_id}")
        index += 8 + chunk_size  # Move to the next chunk

    raise ValueError("WAV file does not contain a 'data' chunk")


def get_audio_length(audio_format: AudioFormat, audio: bytes):
    if audio_format == AudioFormat.WAV_MULAW_8KHZ:
        return len(audio) / 8000.0
    elif audio_format == AudioFormat.OGG_OPUS:
        first_page = OggPage.from_data(audio, 0)
        header = OpusIdPacket.from_ogg_page(first_page)
        page = OggPage.last_page_from_data(audio)
        return page.granule / float(header.input_sample_rate)
    elif audio_format == AudioFormat.MP3:
        return MP3(io.BytesIO(audio)).info.length
    else:
        raise AudioFormatError(
            f"Unsupported audio type '{audio_format}' for getting total audio length"
        )
